#ifndef GTW_GLOB_H
#define GTW_GLOB_H
/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.4                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/******************************************************************************
*******************************************************************************
**
**
     FILE: gtw_glob.h

     DESCRIPTION:  Global declarations and macros for Time Warp Kernel.
        To be included by all kernel and application files.

     LOG HISTORY:
        Changes for incr state saving: new structures for log, Kiran
        Changes for local free pools: Pfree points to local pool, and lock
        on pfree may be redundant
	Added Conservative IO
	Added Optimistic IO2
	Restructuring of software for cloning
*******************************************************************************
******************************************************************************/

/* ---------------------------------------------------------------------------
        INCLUDES
--------------------------------------------------------------------------- */
#include <math.h>
#include <float.h>
#include <limits.h> 
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

#include "gtw_arch.h"

#include "gtw_ds.h"      /* only for using TWds in LPST */
#include "gtw_ds_glob.h" /* only for using TWds in LPST */

#define GTW_MAX_CLONES 64    /* IO4 */

/* uses partitioned pools */
#define partitioned

#define GTW_VERSION " *** Georgia Tech Time Warp (Version 2.7 CL) ***"

#define GTW_MAX_STRING  80
#define GTW_MAX_FILE_NAME_LENGTH 256

/* default distributed computing */
#define DEFAULT_RT_MSG_BUFFER_CHECK 100

/* default period for GVT computation in milliseconds */
#define DEFAULT_GVT_PERIOD 100
#define DEFAULT_MSG_BUFFER_CHECK 100

/* default amount of memory for events per PE (Bytes) */
#define DEFAULT_EV_MEM 1048576
#define DEFAULT_GLOB_EV_MEM 0

/* number consecutive times same GVT computed before error flagged */
#define MAX_SAME_GVT 10000

/* inflation factor beyond min for default # state vectors */
#define DEFAULT_STATE 1.5

/* maximum number of logical processes */
/* .. #define 02/20/99	MAXLP	8000 */
#define	MAXLP	100000 

/* maximum number of physical processors in system */
#define	MAXPROC	 20

/* maximum number of LPs owned by a single PE */
#define MAXLPONPE   (6025)

/* return status from Send */
#define	SUCCEEDED	1

/* #define MAX_PEVENTS_PER_PE 5000 */
#define MAX_PEVENTS_PER_PE 16384

/* large and small numbers */
#define	NEGINFINITY	-FLT_MAX
#define	EPSILON		DBL_EPSILON

/* Processed events marked true, false, in_process or aborted */
#define EXIT_FAIL 	-1

#define FALSE 		0
#define TRUE 		1
#define IN_PROCESS      2
#define	ABORT		3

#define NORMAL          0
#define CALC_GVT        1
#define WAIT_GVT        2

/* Event Types */
#define REGULAR                 0     
#define BLOCKING                1       
#define NONBLOCKING             2                   
#define UNDEFINED               4                
#define OPTIMISTIC		8	             
#define EXTERNAL		9	             
                                                     
/* List Types */
#define INNOLIST                0                    
#define INIOPELIST              1                  
#define INBLOCKINGLIST          2                   
#define INBLOCKINGEVENTLIST     4                 
#define INOPTLIST             	5                
#define INPEXINLIST             6                  
#define INPEXOUTLIST            7                 
                                                   
#define DO_IN_TW                1                  

/* stat flags */
#define STAT_MAPPING   	1
#define STAT_LPS  	2 

/* ---------------------------------------------------------------------------
        MACROS
--------------------------------------------------------------------------- */
#define max(A,B)	((A)>(B) ? (A) : (B))
#define min(A,B)	((A)<(B) ? (A) : (B))

/* LP i resides on PE LP2PE(i) */
#define LP2PE(i)        TWLP[i].Map

/* LP i is the LPinPE(i) th process on that PE */
#define LPinPE(i)       LPNum[i]

/* LPST(i) converts the integer i to a pointer to LP i's  state */
/* #define LPST(i)         (GState[LP2PE(i)].CList[LPinPE(i)]) */

#define LPST(i)  (GState[ TWds.PEToLocalPE[ LP2PE(i) ] ].CList[ LPinPE(i) ])

/* ---------------------------------------------------------------------------
## CANCEL
##
## DESCRIPTION: Cancels events by enqueueing the scheduled event into
##	appropriate CanQ (the events are cancelled by the event that
##	scheduled it originally.)
##-------------------------------------------------------------------------- */
#define CANCEL 								\
	{								\
	NCause = c->NextCause;                         			\
        cpe = LP2PE(c->LP);                            			\
	if ( TWds.MyMachine == TWds.PEToMachine[cpe] )			\
        	cpe = TWds.PEToLocalPE[cpe];                   		\
   	else                                                            \
        	{                                                       \
        	/* Remote - Put in Reflector Thread CanQ */             \
        	cpe = TWnpe;                                		\
		}							\
        if ( GVTflag > 0 ) 						\
		{                                       		\
                if ( MyState->GVT_Progress != WAIT_GVT )              	\
			{						\
                   	(MyState->TransMsgTS)                          	\
                          = min((MyState->TransMsgTS), (c->Msg.RcvTS));	\
                	}                                              	\
                else if ( GVT != MyState->GVT )              		\
			{						\
                   	(MyState->TransMsgTS)                         	\
                       	  = min((MyState->TransMsgTS), (c->Msg.RcvTS));	\
                	}                                             	\
            	}                                                      	\
	/* Finally send the message */                  		\
        TW_LOCK(GState[cpe].Lck_CanQ);                 			\
        c->NextCause = GState[cpe].CanQ;               			\
        GState[cpe].CanQ = c;                          			\
        TW_UNLOCK(GState[cpe].Lck_CanQ);               			\
        TW_POSTSTORE(GState[cpe].Lck_CanQ);				\
	};

/* ---------------------------------------------------------------------------
## MCpSt()
##
## DESCRIPTION: Macro Definition of CpSt
##
##---------------------------------------------------------------------------*/
#define MCpSt(from,to,sz)                               \
{                                                       \
   int i;                                               \
   int nint;                                            \
   int *ifrom, *ito;                                    \
   nint = ((sz) >> 2);                                  \
   ifrom = (int *) (from);                              \
   ito = (int *) (to);                                  \
   for (i = 0; i < nint; i++)                           \
     ito[i] = ifrom[i];                                 \
}


/* ---------------------------------------------------------------------------
	TYPEDEFS
--------------------------------------------------------------------------- */
#if 0
typedef char BOOLEAN;
#endif
typedef int BOOLEAN;

typedef double TWTime;

typedef void (*FuncPtr0) ( void );
typedef void (*FuncPtr1) ( int );
typedef void (*FuncPtr2) ( TW_ADDR_TYPE, int );
typedef void (*FuncPtr ) ( TW_ADDR_TYPE, TW_ADDR_TYPE, int );
typedef void (*CFuncPtr ) ( TW_ADDR_TYPE, TW_ADDR_TYPE, int, int );

typedef void (*IOFuncPtr0) ( void );
typedef void (*IOFuncPtr) ( TW_ADDR_TYPE, TW_ADDR_TYPE, int );

/*****************************************************************************
*                  Kernel Data Structures
*****************************************************************************/

/* ---------------------------------------------------------------------------
** Incremental State Saving Structures
**
** DESCRIPTION: 
**
**---------------------------------------------------------------------------*/
struct log_entry
	{			/* one log */
  	unsigned long addr;
  	unsigned int value;
	};

struct array_ptr
	{ /* link list of pointers */
  	struct log_entry * first, * last; 	/* pointers to array */
  	struct array_ptr * next, * prev; 	/* ptrs to link list elements */
  	TWTime TS;				/* TS corr to first log */
	};

/* ---------------------------------------------------------------------------
** struct EvState
**
** DESCRIPTION: local state
**
**---------------------------------------------------------------------------*/
struct EvState
	{
  	void *state;
  	struct EvState *next;
	};

/* ---------------------------------------------------------------------------
** struct Message
**
** DESCRIPTION: for message
**
**---------------------------------------------------------------------------*/
struct Message
	{
   	TWTime SendTS;			/* time message sent */
   	TWTime RcvTS;			/* time message received */
   	TW_ADDR_TYPE Data;		/* appl. defined data (char*) */
	};


/* ---------------------------------------------------------------------------
** struct TWStats
**
** DESCRIPTION: statistics for logical process
**
**---------------------------------------------------------------------------*/
struct TWStats
   {
   int NEvent;                  /* # events processed */
   int NAbort;                  /* # events aborted */
   int NCancel;                 /* # of processed events cancelled */
   int NStraggler;              /* # straggler induced RBs */
   int NRBStraggler;            /* # proc evnts rolled back by above */
   int NSecondary;              /* # cancellation induced RBs */
   int NRBSecondary;            /* # proc evnts rolled back by above */
   int NArt;                    /* # artificial rollbacks */
   int NRBArt;                  /* # proc events rolled back by above */
   int LongestRB;               /* longest rollback */
   int RemoteMsgSends;          /* Number of Remote +Messages Sent */
   int RemoteAntiMsgSends;      /* Number of Remote AntiMessages Sent */
   TWtimer EvExecTime;
   };

/* ---------------------------------------------------------------------------
** struct Event
**
** DESCRIPTION: Data structure for a single event. This fits in a single 
** 	subpage (128 bytes).  This is important for performance.
**
**---------------------------------------------------------------------------*/
struct Event {
   /* mainstream fields */

   TW_ALIGN BOOLEAN Processed;	/* true if event has been processed */
   BOOLEAN   InPDS;            	/* event in pending event set ?*/
   int LP; 			/* LP on which event is scheduled */

   int SrcLP;			/* source LP */
   int SrcPE;			/* processor originating message */

   struct Message Msg;		/* data field for event */

   struct Event *Next;		/* next event in list for LP */
   struct Event *FNext;		/* free list pointer to next event */
   struct Event *FPrev;		/* free list pointer to previous event */
   struct Event *Prev;		/* previous event */
   struct Event *CauseList;	/* list of events scheduled by this one */
   struct Event *NextCause;	/* next event in causality list */
   struct EvState   *SV;	/* state vector */
   struct log_entry *LogPtr;	/* Last incremental log for this event,
				 * points to the last written log */
   char in;			/* where is the buffer?  */

   int CloneID;			/* IO4 */
   int PCloneID;		/* IO4 */
   int BirthType;		/* IO4 priority in calQ */
   int	   oldLP;		/* IO4 */


#ifdef IODEBUG
  char io_debug[IODEBUGSIZE]; 	/* 1 chars */
  int io_log_ptr;
#endif

  int  EventType;               /*  0=regular, 1=nonblock_io, 3=block_io */
  int  InList;                  /* which list 0=NOLIST  */
  char EvType;      		/* tells you more of type */           

  struct Event *IONext;         /* ptr to next in IoBList */
  struct Event *IOPrev;         /* to next in IoBList */
  IOFuncPtr IOFunc;             /* the event itself or on the LP strct */

  IOFuncPtr RBFunc;       	/* REC: Roll back funtion */
  struct LPState *IoLP;         /* REC: */
};


/* ---------------------------------------------------------------------------
** struct IoBList
**
** DESCRIPTION: Blocked Event List.
**
**---------------------------------------------------------------------------*/
struct IoBListHead 
   {                                                 
   struct Event *IONext;        /* ptr to next Event in EvList*/
   struct Event *IOPrev;        /* ptr to Prev Event in EvList */
   struct LPState *Up;          /* ptr to earlier blocked LP  */
   struct LPState *Down;        /* ptr to later blocked LP  */
   struct Event *LastIoBList;   /* last element io list for LPs */
   };

/* ---------------------------------------------------------------------------
** struct LPState{}
**
** DESCRIPTION: state for a logical process
**
**---------------------------------------------------------------------------*/
struct LPState {
   int ID;			/* number LP (unique) */

   FuncPtr ProcMsg;		/* pointer to regular event routine */
   FuncPtr2 IProcMsg;		/* pointer to initial event routine */
   FuncPtr2 FProcMsg;		/* pointer to final event routine */

   TW_ADDR_TYPE state;		/* events use this area for state */
   TW_ADDR_TYPE auto_chkpt_state;/* start addr of the checkpointable state */
   int CopySize; 		/* bytes to copy for above */
   struct EvState *RBSV;	/* pointer to state to restore after RB */
   TWTime LVT;			/* timestamp of last processed event */

   struct Event EList;		/* event lists for LPs */
   struct Event *Last;		/* last event processed for LP */

   struct IoBListHead IoBList;  /* REC: io list for LPs */

   int NOPEvent, NEXEvent, NNBEvent, NBLEvent;   

   /* aborted **/
   int NOPAbort, NEXAbort, NNBAbort,  NBLAbort;   

   /* number of stragglers */
   int NOPStraggler, NEXStraggler, NNBStraggler;   
   int NBLStraggler, NRBOPStraggler; /* rb b/c of stragglers */

   /* number of cancelled events */
   int NOPCancel;   	
   int NRBOPSecondary;		/* optimistic events rb b/c of cancellation */	
   int NRBDelays;		/* delayed rollbacks */
   int NRBDelaysR;		/* RB done after delay */
   int NEventsBlocked;    
   int NBlockings;    

   /* canned from CQ */
   int NCanOPCQ, NCanEXCQ, NCanNBCQ, NCanBLCQ;

   /* canned from NB */
   int NCanNBQ;	

   /* canned from OQ */
   int NCanOQ;	

   /* canned from BL */
   int NRCanBQ, NBCanBQ;	
   int NBBLEnQ;	/* events in appended to a blocked LP */
   int NBLEnQ;	
   int NOPEnQ;	
   int NNBEnQ;	

   TWtimer BlockExecTime;       /* REC: */
   TWtimer NonBlockExecTime;    /* REC: */
   TWtimer IOExecTime;          /* REC: */

   struct Event *NB_IoEventState; /* REC: pointer to a fossilcollected state */

   /* time warp statistics for logical process */
   int NEvent;		/* # events processed */
   int NAbort;		/* # events aborted */
   int NCancel;		/* # of processed events cancelled */
   int NStraggler;		/* # straggler induced RBs */
   int NRBStraggler;	/* # proc evnts rolled back by above */
   int NSecondary;		/* # cancellation induced RBs */
   int NRBSecondary;		/* # proc evnts rolled back by above */
   int NArt;			/* # artificial rollbacks */
   int NRBArt;			/* # proc events rolled back by above */
   int LongestRB;		/* longest rollback */
   TWtimer EvExecTime;

   /* incremental state save stuff */
   struct array_ptr * Log;	/* start of link list */
   struct log_entry * LogHead;	/* head always points
                                 * to the last location written */
   struct array_ptr * CurArray;	/* array the log head pointer is in */
   BOOLEAN IncrSave;
   BOOLEAN StateValid;		/* true if LP state vector valid */
};


/* ---------------------------------------------------------------------------
** struct RetPool{}
**
** DESCRIPTION: structure for returning buffers to a PE
**
**---------------------------------------------------------------------------*/
struct RetPool {
   TW_ALIGN struct Event *RFree;  /* first buffer to be returned */
   TW_ALIGN struct Event *RLast;  /* last one */
   int NBufs;	 	  /* number of buffers in return pool */
   TW_LOCK_TYPE Lck_RetQ;	  /* mutex lock for accessing the above */
};

/* ---------------------------------------------------------------------------
** struct PEState{}
**
** DESCRIPTION: state for a processor
**
**---------------------------------------------------------------------------*/
struct PEState {
  TW_LOCK_TYPE Free_Lock;	/* lock for free pool */
  TW_ALIGN struct Event *MsgQ;	/* messages sent to this PE */
  TW_LOCK_TYPE Lck_MsgQ;
  TW_ALIGN struct Event *CanQ;	/* messages to be cancelled */
  TW_LOCK_TYPE Lck_CanQ;
  struct EvState *PState; 	/* free pool of states */
  TW_ALIGN struct Event *PFree[MAXPROC];  /* local event free list */
  TW_LOCK_TYPE Lck_PFree;
  struct Event *FLast[MAXPROC];	/* last free event */
  struct RetPool Ret[MAXPROC];
  struct Event *AbortBuf;	/* dummy buffer used by aborted events */

  int NToReturn[MAXPROC];	/* #processed events - #sent to PE */
  int RetTrig[MAXPROC];	/* no. triggering return to sender */
  int NumLocalPool;

  TW_ALIGN int LPsOnPE;	/* number of LPs on this PE */
  TWTime PELVT;        		/* indicates the lower bound on sendts of
				 * messages processed */
  TWTime GVT;			/* latest GVT (init <0) */

  TWTime OptGVT;		/* optimistic GVT IO3 */

  BOOLEAN LVTFlag;		/* set if local LVT computed */
  struct LPState *CList[MAXLPONPE];	/* ptr to info on LPs in this PE */

  struct Event *FirstIoNBList;     /* First Event in IO NB List */
  struct Event *LastIoNBList;      /* Last Event in IO NB List */

  struct LPState *FirstBlockedLP;  /* First Event in IO Blocking List */
  struct LPState *LastBlockedLP;   /* Last Event in IO Blocking List */

  struct Event *FirstIoOptList;     /* IO3: First Event in IO NB List */
  struct Event *LastIoOptList;      /* IO3: Last Event in IO NB List */

  int GVT_Progress;            	/* state of PE w/respect to GVT comp. */
  TWTime  TransMsgTS;          	/* Use by Senders to Keep track of TS  */
  TWtimer StartTime;		/* real time when execution began */

  TWtimer fStartTime;		/* real time when execution began */

  TWtimer EndTime;		/* real time when execution ended */

  int NumFC;		/* no. of FC calls */
  int NumSR;		/* no. of state reclamations */
  int NumTerminate;		/* no. of terminate calls */

  TW_PRIVATE int PEStateMem;	/* Bytes memory per PE for state */
  TW_PRIVATE int randmixer;	/* used to mix random number generator */
  TW_PRIVATE TW_ALIGN TWtimer gvtTimer; /* timer to start GVT computation */

  TW_PRIVATE struct Event *CurEvent;      /* current event that */
  TW_PRIVATE struct LPState *CurState;    /* current LP state */

  TW_PRIVATE barrier_pvt_t Go1; /* for barrier */
  TW_PRIVATE barrier_pvt_t Go2; /* for barrier MHERE added */

  TW_PRIVATE barrier_pvt_t RF1; /* for barrier */
  TW_PRIVATE barrier_pvt_t RF2; /* for barrier */

  /* from memory.c */
  TW_PRIVATE int NumBuf;
  TW_PRIVATE int BufGain;
  TW_PRIVATE int MaxBufGain;
  /* new message buffer gotten by calling TWGetMsg(), init to null */
  TW_PRIVATE struct Event *NewBuf;
  /* message data part seen and typecast by the application */
  TW_PRIVATE void *TWMsg;
  TW_PRIVATE TWtimer repTimer; /* timer for reporting GVT, init to 0 */
  TW_PRIVATE int RepPeriod;	/* period for reporting GVT, init to 1e6 */

  /* from calQ.c */
  TW_PRIVATE struct Event OrigBucket[MAX_PEVENTS_PER_PE];
  TW_PRIVATE struct Event *Bucket;
  TW_PRIVATE int Nbuckets;
  TW_PRIVATE double Width, BucketTop, LastPriority;
  TW_PRIVATE int LastBucket, Qsize, TopThreshold, BottomThreshold;
  TW_PRIVATE char ResizeEnabled;
  TW_PRIVATE int FirstSub;

  int Nret;			/* number of times buffers are returned */
  int Ntaken;			/* buffers are taken */
};


/* ---------------------------------------------------------------------------
** struct TWLPStruct()
**
** DESCRIPTION:  User accessible data structure
**
**---------------------------------------------------------------------------*/
struct TWLPStruct
  {
  int Map;
  FuncPtr2 IProc;
  FuncPtr Proc;
  CFuncPtr CProc;
  FuncPtr2 FProc;

  char *State;
  int  LPStateSize;	

  int CopySize;	

  /* For incremental state save */
  BOOLEAN IncrSave;
  };
typedef struct TWLPStruct TWLPType;


/* ---------------------------------------------------------------------------
## check_macro(ip,statep)
##
## DESCRIPTION:
##
##---------------------------------------------------------------------------*/

#define check_macro(ip,statep) 						\
/* unsigned int * ip;                                    		\
struct LPState * statep; */                              		\
{                                                        		\
  if (statep -> LogHead == statep->CurArray->last){      		\
    if(statep->CurArray->next == NULL){                  		\
      printf("LP %d, time %f\n", statep->ID, statep->Last->Msg.RcvTS);  \
      error("Out of checkpoint space");                   		\
    }                                                    		\
    else{                                                		\
      statep->CurArray = statep->CurArray->next;         		\
      statep->LogHead = statep->CurArray->first;         		\
    }                                                    		\
  }                                                      		\
  else statep->LogHead++;                                		\
                                                         		\
  statep->LogHead->value = (unsigned int) *ip;           		\
  statep->LogHead->addr  = (unsigned long) ip;           		\
}

/* ---------------------------------------------------------------------------
## PUT_FREE_F(ev,pe)
##
## DESCRIPTION:  Adds event 'ev' to front of LOCAL free list of pe.
##      Since it is local no Locks are needed.
##
## TYPE:
##  struct Event *ev; int pe
##
## EXTERNAL VARIABLES:  MyState
##
##---------------------------------------------------------------------------*/
/* struct Event *ev; int pe */
#define PUT_FREE_F(ev,pe) 				\
{                                                       \
  struct Event **_first;				\
  _first = &(MyState->PFree[(pe)]);			\
							\
  (ev)->FNext = *_first;				\
  (ev)->FPrev = NULL;					\
  if (*_first == NULL) {				\
     MyState->FLast[(pe)] = (ev);			\
     *_first = (ev);					\
  }							\
  else {						\
     (*_first)->FPrev = (ev);				\
     *_first = (ev);					\
  }	                                                \
  ev->Next = ev->Prev = NULL;                           \
							\
}

/* ---------------------------------------------------------------------------
## PUT_FREE_E(ev,pe)
##
## DESCRIPTION:  macro to add an event ev to end of LOCAL free list for pe
##
## TYPE:
##  struct Event *ev; int pe
##---------------------------------------------------------------------------*/
/* struct Event *ev; int pe */
#define PUT_FREE_E(ev,pe) 				\
{                                                       \
  struct Event **_last;					\
  _last = &(MyState->FLast[(pe)]);			\
							\
  (ev)->FPrev = *_last;					\
  (ev)->FNext = NULL;					\
  if (*_last == NULL) {					\
     MyState->PFree[(pe)] = (ev);			\
     *_last = (ev);					\
  }							\
  else {						\
     (*_last)->FNext = (ev);				\
     *_last = (ev);					\
  }							\
}

/* ---------------------------------------------------------------------------
## RM_FREE(ev,pe)
##
## DESCRIPTION: removes an event ev from local free list for PE pe
##
## TYPE:
##  struct Event *ev; int pe
##---------------------------------------------------------------------------*/
#define RM_FREE(ev,pe)	 				\
{                                                       \
  if ((ev)->FNext != NULL)				\
     (ev)->FNext->FPrev = (ev)->FPrev;			\
  else /* delete end of free list */			\
     MyState->FLast[(pe)] = (ev)->FPrev;		\
							\
  if ((ev)->FPrev != NULL)				\
     (ev)->FPrev->FNext = (ev)->FNext;			\
  else 	/* delete first element of free list */		\
     MyState->PFree[(pe)] = (ev)->FNext;		\
}

#endif /* GTW_GLOB_H */
