#include "gtw.h"
#include "cs_api.h"     /* cloning api header file */
#include <stdio.h>
#include <string.h>
#include "hashutils.h"

/*********************************************************************************************
Name: addToHashTable

Input Parameters: 
	struct MyStateData* init: Information of the present state and the incoming message.
	struct MyStateData* result: Information of the resultant state and message.
	int index: index into the hash table.
	lpNO: The LP number.

Description: The function adds the  cache structure (struct cacheStruct) in the cache
	     of the LP->lpNO. The structure is added in the index given by the input
	     parameter index. If the slot has already been filled i.e. there are one 
	     or more structures already present in that slot then the structure is 
	     added at the end of the link list for that particular index. If the cache
	     is full then the structure is replaced with the oldest structure in the 
	     link list(The first structure in the link list).

***********************************************************************************************/	
void addToHashTable(struct MyStateData* init, struct MyStateData* result, int index, int lpNO)
{
	struct cacheStruct* temp = NULL ;
	struct cacheStruct* new_struct = NULL ;

	temp = (LPCache[lpNO] + index) ;  /* index in the hashtable */

	/* printf("Adding to Memory--> %d,LP->%d..\n",index,lpNO); */


	if( *(LPFlags[lpNO]+index) != 1 )    /* If the slot is empty */
	{
		memcpy( &(temp->initial), init, sizeof(struct MyStateData) );
                memcpy( &(temp->result), result, sizeof(struct MyStateData) );

		*(LPFlags[lpNO]+index) = 1 ;
		
	}
	else  /*If the slot is filled then add to the link list for that slot */
	{
	    /* Going to the last node */
	    while( temp->next != NULL )
                temp = temp->next ;

	/* Check if cache has not reached the maximum allocated size  */
	    if( (memoryUsed[lpNO] + sizeof(struct cacheStruct)) < maxMemoryPerLP ) 
	    {
                /* printf("CURRENT USED MEMORY -> %d\n",memoryUsed[lpNO]);
		printf("STRUCT SIZE-> %d\n", sizeof(struct cacheStruct));	
 		printf("MAX PER LP->%d\n", maxMemoryPerLP);  */

                /* Getting new memory from the Memory pool. */
		new_struct = (struct cacheStruct*)memAlloc( 1,sizeof(struct cacheStruct));

		memcpy( &(new_struct->initial), init, sizeof(struct MyStateData) );
		memcpy( &(new_struct->result), result, sizeof(struct MyStateData) );

		temp->next = new_struct ;

		memoryUsed[lpNO] = memoryUsed[lpNO] + sizeof(struct cacheStruct) ;
	    }
	    else  
	/* Overwriting existing entry in case of overflow of memory */
	    {
  		/* printf("\nOverwritting existing Memory.") ; */
		memcpy( &(temp->initial), init, sizeof(struct MyStateData) );
                memcpy( &(temp->result), result, sizeof(struct MyStateData) );
	    }

	}

}

/*****************************************************************************************************
Name: checkIfPresent

Input Parameters:
	struct MyStateData* testStructure: Information about present state and incoming message.
	int hashindex: Index into the hash table.
	int lpNO: LP number.

Description: The function checks if similar structure is already present in cache of the particular 
	     LP. The cache is searched at the index given by the input parameter.

Result:
	NULL: If there is no similar entry.
	pointer to the similar structure inside the cache :If a match is found. 

**************************************************************************************************/
struct cacheStruct*  checkIfPresent(struct MyStateData* testStructure, int hashindex, int lpNO)
{
	struct cacheStruct* temp = NULL ;	

	temp = (LPCache[lpNO] + hashindex) ;  /* index in the hashtable */

	memcmp(testStructure, testStructure, sizeof(struct MyStateData) );

	/* If the slot at hashindex is empty return NULL */
	if( *(LPFlags[lpNO]+hashindex) != 1 )
		return NULL ;
	else
	/* Going through the list at hashindex slot */
	{
		do
		{
			/* Comparing memory contents */
			if( memcmp( &(temp->initial),testStructure,sizeof(struct MyStateData) ) == 0 )
			{
				/* printf("FOUND.\n"); */
				noOfFound++ ;
				return temp ;
			}
		}
		while( (temp = temp->next) != NULL ) ;
	}	

	return NULL;	
}


/*******************************************************************************************************
Name: createHashTable

Input parameters:
	int noLP: Number of LP's.

Description: Creates a cache for each LP and initializes few other parameters.

Return Value:
	-1: Unsuccessful.
	 1: Successful.

************************************************************************************************************/
int createHashTable(int noLP)
{
	struct cacheStruct* hashTable = NULL ;
	char* flag = NULL ;
	int i=0 ;

	/* Pointer to the array of pointers. Each member of the array points to cache(Hash table) */
	LPCache   = (struct cacheStruct**)memAlloc(noLP, sizeof(struct cacheStruct*));

	/* Pointer to the array of pointers. Size=Number of LP's */
	LPFlags = (char** )memAlloc(noLP, sizeof(char*)) ; 

	memoryUsed = (int *)memAlloc(noLP, sizeof(int)) ; /* Array used to track memory consumed by each LP. */

	maxMemoryPerLP = (MEMORY_POOL_SIZE-totalMemoryConsumed)/(noLP) - (10 * sizeof(struct cacheStruct) );
        /* printf("MAX MEM PER LP -> %d\n",maxMemoryPerLP) ; */

	for(i=0 ; i < noLP ; i++ )
	{
		/* Allocates initial hash table for each LP */
		hashTable = (struct cacheStruct*)calloc(MAX_HASHVALUE, sizeof(struct cacheStruct));

		LPCache[i] = hashTable ;

		/* Keeps track if the slot is empty or filled.*/
		flag = (char *)memAlloc(MAX_HASHVALUE,sizeof(char)) ;

		LPFlags[i] = flag ;
	}
	
	if( hashTable == NULL || flag == NULL)
		return -1 ;

	return 1 ;

}


/******************************************************************************************
Name: createMemoryPool

Input Parameters:
		None.

Description: Creates Memory pool of the size MEMORY_POOL_SIZE.

Return:

	1:  Successfull
	-1: Unsuccessful.
***********************************************************************************************/
int createMemoryPool()
{
	memoryPool = (char *)calloc(1,MEMORY_POOL_SIZE) ;	
	totalMemoryConsumed = 0 ;

	if( memoryPool == NULL )
		return -1 ;
	memoryPoolPointer = memoryPool ;

	globalMemStruct1 = (struct MyStateData *)calloc(1,sizeof(struct MyStateData) ) ;
        globalMemStruct2 = (struct MyStateData *)calloc(1,sizeof(struct MyStateData) ) ;


	return 1 ;
}


/*********************************************************************************************
Name: memAlloc

Input Parameters:
	int noOfelements: Number of elements.
	int elementSize:  Size of each element.

Description: The function allocates memory equal to noOfelements*elementSize from the memory pool 
	     and updates the current location of memoryPoolPointer.

*************************************************************************************************/
void* memAlloc(int noOfelements, int elementSize)
{
	char* temp = memoryPoolPointer ;

	if( (totalMemoryConsumed + (elementSize * noOfelements)) > MEMORY_POOL_SIZE )
	{
		printf("\nNot enough memory to initialize cache. Exiting ...\n") ;
		exit(0) ;
	}

	memoryPoolPointer = memoryPoolPointer + elementSize * noOfelements ;

	totalMemoryConsumed = totalMemoryConsumed + (elementSize * noOfelements) ;

	return temp ; 
}

