
/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1995 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/

/* ****************************************************************************
**          Memory allocation routines
******************************************************************************/
#include <stdio.h>

#include "gtw_glob.h"
#include "gtw_debug.h"
#include "gtw_utility.h"
#include "gtw_insim.h"
#include "gtw_ds.h"

/* GLOBAL */
TW_SHARED int TWMemMap[D_MAXPROC*MAXMACHINES][D_MAXPROC*MAXMACHINES];

/* LOCAL */
static int nsp_ev, nsp_msg;	/* no. of subpages in event, message */

/* forward references */
void AllocateEventMemory( int, int, int );
void PrintBufAlloc( void );
void AllocateStateVector( TW_LONG, int );
void AllocateLPState( int );
char * TWMalloc( TW_LONG );
void _TWAutoCheck( char *, int, int );
void InitMemMap( void );
void InitPEXMemMap ( int );

void AllocateLPState( int );
void AllocatePEXMemory ( int , int);

extern int show_noprint; /* IO4 */

int mh_alloc[D_MAXPROC] = {0};
/* int mh_iproc_alloc[D_MAXPROC] = {0}; */

/*----------------------------------------------------------------------------
**  MemAllocate (int, TW_ADDR_TYPE, TW_ADDR_TYPE, int )
**
**  DESCRIPTION:
**  	Allocate a chunk of memory at least pebytes for a
**	single PE, first_addr returns pointer to first address
**	of block, last_addr returns pointer to last one.
**
**---------------------------------------------------------------------------*/
void MemAllocate (
		int pebytes, 			/* memory per PE in bytes */
		TW_ADDR_TYPE * first_addr,
		TW_ADDR_TYPE * last_addr,
		int  MyPE
		)
{
   int npages, size, evsize;
   TW_ADDR_TYPE addr;		/* starting addr. of allocated memory */

   if( pebytes%PG_SIZE == 0 )
	npages = pebytes / PG_SIZE;
   else
      	npages = (pebytes / PG_SIZE) + 1;

   npages = npages + 1; 		/* one extra for alignment */
   size = npages * PG_SIZE;

#if 0
   printf("\nAllocating %d bytes for events for PE %d ppg size=%d "
        "spg size=%d nsp_ev=%d\n",
                size, MyPE, PG_SIZE, SP_SIZE, nsp_ev );  /* IO5 */
#endif

   if((addr = (char *) calloc(1,(unsigned) size)) == NULL)
	error("PE %d, calloc unable to allocate memory for events", MyPE);
   /* MH_HERE */ mh_alloc[MyPE] += size;

   *last_addr = addr + size - 1;
   /* get alignment right */
   if(((unsigned long) addr % PG_SIZE) == 0)
      	*first_addr = addr;
   else
      	*first_addr = addr + PG_SIZE - ((unsigned long) addr % PG_SIZE);

   if( MyPE == 0 ) 
	{
     	if( !(TWdebug & DEBUG_SILENT) && !(show_noprint)  )
		{
   		evsize = (nsp_ev + nsp_msg) * SP_SIZE;
      		printf("\n PE %d: %d bytes per event, %d bytes per message\n",
	   	   MyPE, nsp_ev*SP_SIZE, nsp_msg*SP_SIZE);
      		printf(" Allocate = %8.3f MB (%d pages or %d buffers) ... \n",
	   	 (float)((long)*last_addr - (long)*first_addr + 1) / 1000000.0,
	   	   ((long)*last_addr - (long)*first_addr + 1) / PG_SIZE,
	   	   ((long)*last_addr - (long)*first_addr + 1)/evsize);
     		fflush(stdout);
    		}
   	}
}


/*----------------------------------------------------------------------------
**  AllocateEventMemory ()
**
**  DESCRIPTION:
** 	Allocate memory for free list of events and set it up 
**
**---------------------------------------------------------------------------*/
void AllocateEventMemory
		(
		int MemPerPE, 	/* amount to allocate per PE in bytes */
		int GMemPerPE,  /* amount for the global pool per PE */
		int MyPE
		)
{
   TW_LONG i, j, k, cnt, BufsPerPE, nbufs, npages, BufsPerPage;
   TW_LONG TotalPEMem;
   TW_ADDR_TYPE first, last, ev_pool_Free;
   struct Event *prev, *EFree;

   /* pad TWMsgSize so that they can be copied integer by integer */

   if(TWMsgSize < sizeof(TW_LONG))
      	TWMsgSize = sizeof(TW_LONG);
   if((TWMsgSize % sizeof(TW_LONG)) != 0)
      	TWMsgSize = ((int) (TWMsgSize / sizeof(TW_LONG)) + 1) * sizeof(TW_LONG);

   nsp_ev = (((int) sizeof(struct Event) % SP_SIZE) == 0) ? 
      	(int) sizeof(struct Event) / SP_SIZE :
      	(int) sizeof(struct Event) / SP_SIZE + 1;
   nsp_msg = (((int) TWMsgSize % SP_SIZE) == 0) ?
      	(int) TWMsgSize / SP_SIZE :
      	(int) TWMsgSize / SP_SIZE + 1;

   MemAllocate(MemPerPE, &first, &last, MyPE);
   if(((long)last - (long)first + 1) < MemPerPE)
      	warn("PE %d only allocated %d bytes for events (%d requested)\n",
		MyPE, (long) last - (long)first + 1, MemPerPE);

   BufsPerPE = ((long)last - (long)first + 1) / ((nsp_ev + nsp_msg) * SP_SIZE);

   /* Set up buffer pools in each PE */
   ev_pool_Free = first;
   cnt = 0;
   TotalPEMem = 0;

   for(j=0; j<TWnpe+RFthread; j++) 
	TotalPEMem += TWMemMap[MyPE][j];

   if(TotalPEMem==0) 
	error ("PE %d has no buffer memory in TWMemMap\n", MyPE);

   for(j=0 ; j<TWnpe+RFthread; j++) 
	{
      	nbufs = (BufsPerPE * TWMemMap[MyPE][j]) / TotalPEMem;
      	if((nbufs == 0) && (TWMemMap[MyPE][j] != 0))
		{
	  	printf("BufsPerPE %d, TotalPEMem %d, TWMemMap[][]=%d\n", 
			BufsPerPE, TotalPEMem, TWMemMap[MyPE][j]);
	 	warn ("TWMemMap[%d][%d] non-zero, but no buffers allocated\n",
			MyPE,j);
		}

      	if(nbufs == 0) 
		{
         	GState[MyPE].PFree[j] = NULL;
         	GState[MyPE].FLast[j] = NULL;

         	GState[j].NToReturn[MyPE] = 0;	  /* count of remote buffers */
	 	GState[j].RetTrig[MyPE] = 0;  /* threshold triggering return */

	 	continue;
      		}

      	GState[MyPE].PFree[j] = (struct Event *) ev_pool_Free;
      	prev = NULL;

      	GState[j].NToReturn[MyPE] = 0;	  /* count of remote buffers */
      	GState[j].RetTrig[MyPE] = nbufs/RetThreshold;

      	/* threshold triggering return */

	for(k=0; k<nbufs; k++) 
		{
		EFree = (struct Event *) ev_pool_Free;
         	EFree->FPrev = prev;
         	prev = EFree;
         	ev_pool_Free += nsp_ev * SP_SIZE;
         	EFree->Msg.Data = ev_pool_Free;
         	ev_pool_Free += nsp_msg * SP_SIZE;
         	EFree->FNext = (struct Event *) ev_pool_Free;
         	EFree->SV = NULL;
	 	EFree->LP = 0;		/* need for initial buffer allocation */
	 	EFree->InPDS = FALSE;	
         	EFree->Msg.RcvTS = NEGINFINITY;
         	cnt++;
      		}

      	if(ev_pool_Free > last+1)
         	error ("Internal memory allocation error\n");
      	GState[MyPE].FLast[j] = EFree;
      	EFree->FNext = NULL;
   	}

   if(cnt > BufsPerPE)
      	error ("Problem in event memory allocation\n");

   /* allocate additional buffers to global pool, in 1 page blocks */
   /* round up to nearest page and allocate memory */
   cnt = 0;
   if( GMemPerPE > 0 ) 
	{
      	MemAllocate(GMemPerPE, &first, &last, MyPE);
      	npages = ((long)last - (long)first + 1) / PG_SIZE;
      	PFree = NULL;
      	BufsPerPage = PG_SIZE / ((nsp_ev + nsp_msg) * SP_SIZE);
      	if(BufsPerPage == 0) 
		{
	 	warn ("PE %d: Message size too large for global pool\n", MyPE);
      		} 
	else 
		{
		for(i=0; i<npages; i++) 
			{
            		/* link BLOCKS of global pool buffers w/ Next field */
	    		ev_pool_Free = first;
	    		first += PG_SIZE;
            		EFree = (struct Event *) ev_pool_Free;
            		EFree->Next = PFree;
            		PFree = EFree;
            		prev = NULL;
            		for(j=0; j<BufsPerPage; j++) 
				{
				EFree = (struct Event *) ev_pool_Free;
	       			EFree->FPrev = prev;
	       			prev = EFree;
               			ev_pool_Free += nsp_ev * SP_SIZE;
               			EFree->Msg.Data = ev_pool_Free;
               			ev_pool_Free += nsp_msg * SP_SIZE;
               			EFree->FNext = (struct Event *) ev_pool_Free;
               			EFree->SV = NULL;

				/* need for initial buffer allocation */
	       			EFree->LP = 0;		
	       			EFree->Msg.RcvTS = NEGINFINITY;
	       			cnt++;
            			}
            		EFree->FNext = NULL;
	 		}  /* outer for loop - for each page */
      		} /* else */
   	} /* if global pool buffers to allocate */
   NumGlobalPool = cnt;
}


/*----------------------------------------------------------------------------
**  PrintBufAlloc ()
**
**  DESCRIPTION:
** 	Print out buffer allocation
**
**---------------------------------------------------------------------------*/
void PrintBufAlloc( void )
{  TW_LONG i, j, n;
   struct Event *p;

   printf ("BUFFER ALLOCATION (in K Buffers):\n");
   printf("PE ");
   for(i=0; i<TWnpe+RFthread; i++) printf (" %2d ", i);
   printf("\n");
   for(i=0; i<TWnpe+RFthread; i++) {
      printf ("%2d:", i);
      for(j=0; j<TWnpe+RFthread; j++) {
         n=0;
         for(p=GState[i].PFree[j]; p!=NULL; p=p->FNext) n++;
         /* printf (" %3.1f", n/1000.0); */
	 printf("%4d ", n);
      }
      printf ("\n");
   }
}


/*----------------------------------------------------------------------------
**  AllocateStateVector ()
**
**  DESCRIPTION:
**  	Allocate "pemem" byte of memory for state vectors for LPs.
**  	Some are attached to the event buffers and some are unattached.
**  	This procedure is called by each PE after the application IProc
**  	procedures have been called.
**
**  	State vectors never migrate across processor boundaries.
**
**---------------------------------------------------------------------------*/
void
AllocateStateVector(TW_LONG pemem, int MyPE)
{
   TW_LONG i, nsv, minsize;
   struct Event *p;
   struct EvState *s;
   char *stateptr, *lastaddr;

   /* figure out state vector size */
   MaxStateSize = 0;
   for(i = 0; i < TWnlp; i++) /* IO4 */
	{
	if( RFthread )
      	    MaxStateSize = max ( MaxStateSize, TWLP[i].CopySize );
	else
      	    MaxStateSize = max ( MaxStateSize, LPST(i)->CopySize );
	}

   /* pad MaxStateSize to ingegral multiple of integer size */
   if(MaxStateSize < sizeof(TW_LONG))
      	MaxStateSize = sizeof(TW_LONG);
   if((MaxStateSize % sizeof(TW_LONG)) != 0)
      	MaxStateSize 
	    = ((int) (MaxStateSize / sizeof(TW_LONG)) + 1) * sizeof(TW_LONG);

   /* figure out how many state vectors are needed first count free pools */
   nsv = 0;
   for(i=0; i<TWnpe+RFthread; i++)
      	for(p=GState[MyPE].PFree[i]; p!=NULL; p=p->FNext) nsv++;

   /* add in initial (dummy) buffers stored in event lists */
   for(i=0; i<GState[MyPE].LPsOnPE; i++)
      	for(p=&(GState[MyPE].CList[i]->EList); p!=NULL; p=p->Next) nsv++;

   /* check initial buffers stored in MsgQ */
   for(p=GState[MyPE].MsgQ; p!=NULL; p=p->Next) nsv++;

   /* add in buffers in calendar queue (scheduled by IProc) */
   nsv = nsv + GState[MyPE].Qsize;

   /*
    * printf ("\nPE %d, %d state vectors, memory=%d bytes, %d bytes per SV," 
    * " MaxStateSize=%d\n",
    * MyPE, nsv, pemem, sizeof(struct EvState) + MaxStateSize, MaxStateSize);
    * fflush(stdout);
    */

   minsize = (sizeof(struct EvState) + MaxStateSize) * (DEFAULT_STATE * nsv);

   if(pemem < 0) pemem = minsize;	/* default */
   if(pemem < minsize) 
	{
      	warn("Min state memory of 1 per event (%d bytes) being used\n", 
		minsize);
      	pemem = minsize;
   	}

   /* allocate the memory */
   GState[MyPE].PEStateMem = pemem;
   if((stateptr = (char *) calloc(1,pemem)) == NULL)
      	error("Unable to allocate state vector memory");
   lastaddr = stateptr + pemem;
   /* MH_HERE */ mh_alloc[MyPE] += pemem;

   /* attach states to events */

   /* skip events in free list; will probably be stripped */

   /* events in event lists */
   for(i=0; i<GState[MyPE].LPsOnPE; i++) 
	{
      	for(p=&(GState[MyPE].CList[i]->EList); p!=NULL; p=p->Next) 
		{
	 	p->SV = (struct EvState *) stateptr;
	 	stateptr = stateptr + sizeof (struct EvState);
         	p->SV->state = stateptr;
         	p->SV->next = NULL;
         	stateptr = stateptr + MaxStateSize;
      		}
   	}

   /* initial events stored in MsgQ */
   for(p=GState[MyPE].MsgQ; p!=NULL; p=p->Next) 
	{
      	p->SV = (struct EvState *) stateptr;
      	stateptr = stateptr + sizeof (struct EvState);
      	p->SV->state = stateptr;
      	p->SV->next = NULL;
      	stateptr = stateptr + MaxStateSize;
   	}

   /*
    * printf("Allocate extra state vector, stateptr=%x, lastaddr=%x\n",
    * stateptr,lastaddr);
    */

   GState[MyPE].PState = NULL;

   /* put remaining into state queue free list */
   while(1) 
	{
      	if((stateptr+sizeof(struct EvState)+MaxStateSize) > lastaddr) break;
      	s = (struct EvState *) stateptr;
      	s->next = GState[MyPE].PState;
      	GState[MyPE].PState = s;
      	stateptr = stateptr + sizeof (struct EvState);
      	s->state = stateptr;
      	stateptr = stateptr + MaxStateSize;
   	}

}


/*----------------------------------------------------------------------------
**  AllocateLPState ()
**
**  DESCRIPTION:
** 	Allocate memory for LPStates
**
**---------------------------------------------------------------------------*/
void
AllocateLPState(int pe)
{
   int i;
   for(i = 0; i < GState[pe].LPsOnPE; i++) 
	{
      	if((GState[pe].CList[i]
	   = (struct LPState *) calloc(1,sizeof(struct LPState))) == NULL)
		error("LPState callocation problem");
	/* MH_HERE */ mh_alloc[pe] += sizeof(struct LPState);
   	}
}


/*----------------------------------------------------------------------------
**  TWMalloc ()
**
**  DESCRIPTION:
** 	called from IProc()
**
**---------------------------------------------------------------------------*/
char * TWMalloc( TW_LONG size )
{
   char *p;

   if(size <= 0)
      	return (NULL);

   if(size < sizeof(TW_LONG))
      	size = sizeof(TW_LONG);
   if((size % sizeof(TW_LONG)) != 0)
      	size = ((int) (size / sizeof(TW_LONG)) + 1) * sizeof(TW_LONG);

   if((p = (char *) calloc(1,size)) == NULL) 
	{
      	printf("WARNING: TWMalloc() couldn't allocate memory \n");
      	return(NULL);
   	}
   /* MH_HERE  mh_iproc_alloc[MyPE] += size; */
   return (p);
}


/*----------------------------------------------------------------------------
**  _TWAutoCheck ()
**
**  DESCRIPTION:
** 	called from IProc(), sets auto_chkpt_state to point to beginning of
**	state that is to be automatically checkpointed.
**	Also sets the copy size of this checkpointed state.
**
**---------------------------------------------------------------------------*/
void 
_TWAutoCheck(char *ptr, int size, int MyPE)
{
   GState[MyPE].CurState->auto_chkpt_state = ptr;
   GState[MyPE].CurState->CopySize = size;
}


/*----------------------------------------------------------------------------
**  InitMemMap ()
**
**  DESCRIPTION:
**	Initialize TWMemMap array
**
**---------------------------------------------------------------------------*/
void 
InitMemMap(void)
{
   int i, j;

   for(i=0; i<TWnpe+RFthread; i++)
      	for(j=0; j<TWnpe+RFthread; j++)
	 	TWMemMap[i][j] = 0;
}

