/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/


/****************************************************************************
*
*                Calendar Q for pending event set of TW
*                One such Q per processor. Rolled back events
*                are reinserted in the calendar queue.
*
*     Almost identical to the pseudocode in Brown's
*     CACM paper. Initially implemented in C++ by Chris 
*     Carothers. Modified for Time Warp.  -- SRD, Oct 1992.
*
*     Added heuristics to switch off resizing once queue 
*     capacity is exceeded -- SRD, May 1993. 
*
*     Added routine Delete_Arbitrary() for deletion of
*     arbitrary elements -- SRD, Nov 1993
*
*     Added routine Reclaim_future_from_PDS() for cancelback
*     -- SRD Aug 1994 
*     Added priority for cloning events
*     -- MH 1998
*
*****************************************************************************/

#include <stdio.h>
#include "gtw_glob.h"
#include "gtw_debug.h"
#include "gtw_utility.h"
#include "gtw_insim.h"


/* GLOBALS  */
int largestQsize[20]={0};

/* LOCAL prototypes */
struct Event 	*Delete(int, int);
void 		Delete_Arbitrary( struct Event **, int );
struct Event 	*DeQueue(int);
void 		DumpBucket(int);
void 		EnQueue(struct Event *, int);
void 		Insert(int, struct Event *, int);
void 		LocalInit(int , int, double, double, int);
double 		Minim_Q(int);
double 		NewWidth(int);
void 		ReSize(int, int);
double 		Smallest(int);


/* ---------------------------------------------------------------------------
**  DeQueue()
**
**  DESCRIPTION:
**
** DONE
**---------------------------------------------------------------------------*/
struct Event * DeQueue(int MyPE)
{
   int i;
   int smallest;
   struct Event *E;
   double tmp;

   struct Event * bucket;	/* pointer to Calender queue bucket */
   struct PEState * p_ptr; 

   p_ptr = & GState[MyPE];

   if(p_ptr->Qsize == 0)
	{
     	return (NULL);
   	}

   p_ptr = &GState[MyPE];
   bucket = p_ptr->Bucket;

   for(i = p_ptr->LastBucket;;)
	{
      	if(  (bucket[i].Next != NULL)
	     && (bucket[i].Next->Msg.RcvTS < p_ptr->BucketTop))
		{
	 	p_ptr->LastBucket = i;
	 	p_ptr->LastPriority = bucket[i].Next->Msg.RcvTS;
	 	E = Delete(i,MyPE);
		p_ptr->Qsize--;
	 
	 	if(p_ptr->Qsize < p_ptr->BottomThreshold)
	   		ReSize(p_ptr->Nbuckets / 2, MyPE);
	 	/* return (E); */
	 	goto END;
      		} 
	else 
		{
	 	++i;
	 	if(i == p_ptr->Nbuckets)
	    		i = 0;
	 	p_ptr->BucketTop = p_ptr->BucketTop + p_ptr->Width;
	 	if(i == p_ptr->LastBucket)
	    		break;
      		}
   	}

   /* Start Direct Search */
   for(smallest = 0; bucket[smallest].Next == NULL; smallest++)
      	if(smallest >= p_ptr->Nbuckets)
		{
	 	printf("Nbuckets = %d, p_ptr->Qsize = %d \n",
			p_ptr->Nbuckets, p_ptr->Qsize);
	 	error("Grave inconsistency in calendar queue");
      		}

   for(i = smallest + 1; i < p_ptr->Nbuckets; i++)
	 {
      	if(bucket[i].Next == NULL)
	 	continue;
      	else if(bucket[i].Next->Msg.RcvTS < bucket[smallest].Next->Msg.RcvTS)
	 	smallest = i;
   	}

   (p_ptr->Qsize)--;
   E = Delete(smallest,MyPE);
   p_ptr->LastBucket = smallest;
   p_ptr->LastPriority = E->Msg.RcvTS;
   tmp = p_ptr->Width * 0.5;
   p_ptr->BucketTop = p_ptr->Width
	* (double) ((int) (p_ptr->LastPriority / p_ptr->Width) + 1);
   p_ptr->BucketTop = p_ptr->BucketTop + tmp;


END:
   /* SANITY CHECK */
   if(TWdebug & DEBUG_CALQ)
     	if(Smallest(MyPE) < E->Msg.RcvTS)
		{
       		printf("PE:%d, smallest = %f, got %f\n",MyPE, Smallest(MyPE),
	      		E->Msg.RcvTS);
       		DumpBucket(MyPE);
       		fflush(stdout); 
       		quit(MyPE);
     		} 
   return (E);
}


/* ---------------------------------------------------------------------------
**  Delete()
**
**  DESCRIPTION:
**	deletes first element in Bucket[index]
**  	Changes Bucket[index] 
**
** DONE
**---------------------------------------------------------------------------*/
struct Event * Delete(int index, int MyPE)
{
   struct Event *temp;

   struct PEState * p_ptr;
   struct Event * bucket;
   p_ptr = &GState[MyPE];

   bucket = p_ptr->Bucket;

   temp = bucket[index].Next;
   if(bucket[index].Next != NULL)
	{
     	bucket[index].Next = bucket[index].Next->Next;
     	if(bucket[index].Next != NULL)
       		bucket[index].Next->Prev = & bucket[index];
   	}

   temp->Next = NULL;
   temp->Prev = NULL;
   temp->InPDS = FALSE;

   /* no need to restore since data structure in GState has changed */
   return (temp);
}


/* ---------------------------------------------------------------------------
**  EnQueue()
**
**  DESCRIPTION:
**
**  DONE
**---------------------------------------------------------------------------*/
void
EnQueue(struct Event * NewE, int MyPE)
{
   int i;
   double tmp;

   struct PEState * p_ptr;
   p_ptr = & GState[MyPE];

   i = (int) (NewE->Msg.RcvTS / p_ptr->Width);
   i = i % p_ptr->Nbuckets;
   Insert(i, NewE, MyPE);
   ++(p_ptr->Qsize);

   /* Code for Out of Sequence EnQueues and DeQueues.
    * Required for insertions afetr a rollback */

   if(NewE->Msg.RcvTS <= p_ptr->LastPriority)
	{
      	tmp = p_ptr->Width * 0.5;
      	p_ptr->LastPriority = NewE->Msg.RcvTS;
      	p_ptr->BucketTop = p_ptr->Width
	  * (double) ((int) (p_ptr->LastPriority / p_ptr->Width) + 1);
      	p_ptr->BucketTop = p_ptr->BucketTop + tmp;
      	p_ptr->LastBucket = i;
   	}

   if(p_ptr->Qsize > p_ptr->TopThreshold)
	ReSize(2 * p_ptr->Nbuckets, MyPE);
}


/* ---------------------------------------------------------------------------
**  Insert()
**
**  DESCRIPTION:
**
**  DONE
**---------------------------------------------------------------------------*/
void
Insert(int bucketindex, struct Event * NewE, int MyPE)
{
   struct Event *p;
   struct Event *bucket;

   struct Event *s;

   bucket = &(GState[MyPE].Bucket[0]);
   
   if( bucket[bucketindex].Next == NULL )
	{	/* bucket empty */
      	bucket[bucketindex].Next = NewE;
      	NewE->Next = NULL;
      	NewE->Prev = &bucket[bucketindex];
   	} 
   else if( bucket[bucketindex].Next->Msg.RcvTS > NewE->Msg.RcvTS) 
	{
      	/* insert at the front of this bucket */
      	NewE->Next = bucket[bucketindex].Next;
      	NewE->Next->Prev = NewE;
      	NewE->Prev = &bucket[bucketindex];
      	bucket[bucketindex].Next = NewE;
   	} 
   else if((bucket[bucketindex].Next->Msg.RcvTS == NewE->Msg.RcvTS)
	&& (NewE->BirthType == -2) )  /* IO4 */
	{
      	/* insert at the front of this bucket */
      	NewE->Next = bucket[bucketindex].Next;
      	NewE->Next->Prev = NewE;
      	NewE->Prev = &bucket[bucketindex];
      	bucket[bucketindex].Next = NewE;
	}
   else 
	{
      	/* insert somewhere else at the end */
      	for(p = bucket[bucketindex].Next; p->Next != NULL; p = p->Next)
		{
		if( NewE->BirthType == -2 )  /* IO4 */
			{
	 		if(  (p->Msg.RcvTS < NewE->Msg.RcvTS) 
				&& (p->Next->Msg.RcvTS >= NewE->Msg.RcvTS))
			break;
			}
	 	else if(  (p->Msg.RcvTS <= NewE->Msg.RcvTS) 
			&& (p->Next->Msg.RcvTS > NewE->Msg.RcvTS))
			{
	    		break;
			}
      		}

      	NewE->Next = p->Next;
      	if(p->Next != NULL)
	 	p->Next->Prev = NewE;
      	p->Next = NewE;
      	NewE->Prev = p;
   	}

   NewE->InPDS = TRUE;
   NewE->in = 'c';
}


/* ---------------------------------------------------------------------------
**  LocalInit()
**
**  DESCRIPTION:
**
** DONE
**---------------------------------------------------------------------------*/
void
LocalInit(int qbase, int nbuck, double bwidth, double startprio, int MyPE)
{
   int i;
   int n;
   double tmp;

   struct PEState * p_ptr;
   p_ptr =  &GState[MyPE];
   /* Local variables */

   p_ptr -> FirstSub = qbase;
   p_ptr -> Bucket = &(p_ptr -> OrigBucket[p_ptr->FirstSub]);
   p_ptr -> Width = bwidth;
   p_ptr -> Nbuckets = nbuck;
   p_ptr -> Qsize = 0;

   for(i = 0; i < nbuck; i++)
     	p_ptr -> Bucket[i].Next = NULL;

   p_ptr -> LastPriority = startprio;
   n = (int) startprio / bwidth;
   p_ptr -> LastBucket = (n % nbuck);
   p_ptr -> BucketTop = (bwidth * (double) (n + 1));
   tmp = bwidth * 0.5;
   p_ptr -> BucketTop = p_ptr -> BucketTop + tmp;
   p_ptr -> BottomThreshold = nbuck / 2 - 2;
   p_ptr -> TopThreshold = 2 * nbuck;
}


/* ---------------------------------------------------------------------------
**  NewWidth()
**
**  DESCRIPTION:
**	does not change queue, however, does dequeue nsamples and
**	enqueues them back again 
** DONE
**---------------------------------------------------------------------------*/
double NewWidth(int MyPE)
{
   int i, nsamples, lastbucket, count;
   double AVG[25], avg1, avg2, buckettop, lastprio;
   struct Event *E[25];

   struct PEState * p_ptr;
   p_ptr = & GState[MyPE];

   avg1 = avg2 = 0.0;

   if(p_ptr->Qsize < 2) 
	{
     	return (1.0);
   	} 
   else if(p_ptr->Qsize <= 5)
      	nsamples = p_ptr->Qsize;
   else
      	nsamples = 5 + (p_ptr->Qsize / 10);

   if(nsamples > 25)
      nsamples = 25;

   p_ptr->ResizeEnabled = FALSE;
   buckettop = p_ptr->BucketTop;
   lastprio = p_ptr->LastPriority;
   lastbucket = p_ptr->LastBucket;

   for(i = 0; i < nsamples; i++)
   	E[i] = DeQueue(MyPE);

   for(i = 0; i < nsamples - 1; i++)
	{
      	AVG[i] = E[i + 1]->Msg.RcvTS - E[i]->Msg.RcvTS;
      	avg1 = avg1 + AVG[i];
   	}

   for(i = 0; i < nsamples; i++)
	{
      	EnQueue(E[i], MyPE);
   	}

   p_ptr->ResizeEnabled = TRUE;
   p_ptr->BucketTop = buckettop;
   p_ptr->LastPriority = lastprio;
   p_ptr->LastBucket = lastbucket;

   avg1 = avg1 / (double) (nsamples - 1);
   avg1 = avg1 * 2.0;

   count = 0;
   for( i = 0; i < nsamples - 1; i++)
      	if(AVG[i] < avg1) 
		{
	 	avg2 = avg2 + AVG[i];
	 	count++;
      		}

   if((count == 0) || (avg2 == 0))
	{
     	return (1.0);
   	}

   avg2 = avg2 / (double) count;
   avg2 = avg2 * 3.0;
   
   return (avg2);
}

/* ---------------------------------------------------------------------------
**  ReSize()
**
**  DESCRIPTION:
**	Resize computes a new calendar queue size, and copies the old calendar
**  	queue into the new one. It is called when there are too many elements
**	in the calendar queue
**
** DONE
**---------------------------------------------------------------------------*/
void ReSize(int newsize, int MyPE)
{
   double bwidth;
   int i;
   int oldnbuckets;
   struct Event *oldbucket;
   struct Event *temp;

   static int warning_issued = 0; /* MARIA ADDED */

   struct PEState * p_ptr;
   p_ptr = & GState[MyPE];


   if(!(p_ptr->ResizeEnabled))
	{
     	return;
   	}

   if(p_ptr->Nbuckets + newsize >= MAX_PEVENTS_PER_PE) 
	{
     	if( warning_issued == 0 ) /* MARIA ADDED */
         	{
         	warning_issued = 1;
         	printf("Calendar Q capacity exceeded on PE %d\n", MyPE);
                printf("\tENLARGEMENT (%d, %d)\n",
                MAX_PEVENTS_PER_PE, p_ptr->Nbuckets + newsize );
		}
      	if(TWdebug & DEBUG_CALQ)
		printf("Calendar Q capacity exceeded on PE %d\n", MyPE);
      	return;
   	}

   if( largestQsize[MyPE] < ( p_ptr->Nbuckets + newsize ) ) /*MARIA*/
        largestQsize[MyPE] = p_ptr->Nbuckets + newsize;

   bwidth = NewWidth(MyPE);
   if(TWdebug & DEBUG_CALQ)
	{
      	printf("New Width = %f, New Size = %d\n", bwidth, newsize);
      	fflush(stdout);
   	}

   oldbucket = p_ptr->Bucket;
   oldnbuckets = p_ptr->Nbuckets;

   if(p_ptr->FirstSub == 0)
     	LocalInit(MAX_PEVENTS_PER_PE - newsize, newsize, bwidth,
		p_ptr->LastPriority, MyPE);
   else
      	LocalInit(0, newsize, bwidth, p_ptr->LastPriority,MyPE);

   for(i = 0; i < oldnbuckets; i++) 
	{
      	while(oldbucket[i].Next != NULL)
		{
	 	temp = oldbucket[i].Next;
	 	oldbucket[i].Next = temp->Next;
	 	if(temp->Next != NULL)
	    		temp->Next->Prev = &oldbucket[i];
	 	temp->Next = NULL;
	 	temp->Prev = NULL;
	 	EnQueue(temp, MyPE);
      		}
   	}
}

/* ---------------------------------------------------------------------------
**  DumpBucket()
**
**  DESCRIPTION:
**	 print calendar queue by going thru the buckets in turn
**
** DONE
**---------------------------------------------------------------------------*/
void DumpBucket(int MyPE)
{
   int i;
   struct Event *probe;
   extern TW_SHARED TW_LOCK_TYPE dumplck;	/* for dumping */

   struct PEState * p_ptr;
   p_ptr = & GState[MyPE];
   
   TW_LOCK(dumplck);
   printf("*----------- Dumping Buckets, PE %d--------------*\n", MyPE);
   for(i = 0; i < p_ptr->Nbuckets; i++) 
	{
	probe = p_ptr->Bucket[i].Next;
      	printf("Bucket[%d]: ", i);
      	while(1) 
		{
		if(probe == NULL)
			{
	    		printf(" NULL\n");
	    		break;
	 		} 
		else 
			{
	    		printf(" %5.3f-%d", probe->Msg.RcvTS, probe->Processed);
	   		probe = probe->Next;
	 		}
      		}
   	}

   printf("*----------- Finished Dumping --------------*\n");
   TW_UNLOCK(dumplck);
}


/* ---------------------------------------------------------------------------
**  InitQ()
**
**  DESCRIPTION:
** 	intialize queue to be with a size of 2, and a width of 1.0 per bucket
**
**  DONE
**---------------------------------------------------------------------------*/
void InitQ(int MyPE)
{
   LocalInit(0, 2, 1.0, 0.0, MyPE);
   GState[MyPE].ResizeEnabled = TRUE;
}


/* ---------------------------------------------------------------------------
**  Smallest()
**
**  DESCRIPTION:
**	Find the smallest element in the calq. Useful for debugging
**
** DONE
**---------------------------------------------------------------------------*/
double Smallest(int MyPE)
{
   int i;
   struct Event *probe;
   double smallest = MAXFLOAT;
   int cnt = 0;

   struct PEState * p_ptr;
   p_ptr = & GState[MyPE];

   for(i = 0; i < p_ptr->Nbuckets; i++) 
	{
      	probe = p_ptr->Bucket[i].Next;
      	while(1) 
		{
	 	if(probe == NULL) 
			{
	    		break;
	 		} 
		else 
			{
	    		if(probe->Msg.RcvTS < smallest)
	       			smallest = probe->Msg.RcvTS;
	   		probe = probe->Next;
	    		cnt++;
	 		}
      		}
   	}

   if(cnt != p_ptr->Qsize)
      	printf("Wrong Qsize, Qsize=%d, found %d\n", p_ptr->Qsize, cnt);
   fflush(stdout);
   
   return (smallest);
}

/* ---------------------------------------------------------------------------
**  Reclaim_future_from_PDS()
**
**  DESCRIPTION:
**
** DONE
**---------------------------------------------------------------------------*/
void Reclaim_future_from_PDS(TWTime ts, int MyPE)
{
   ts = ts;
   MyPE = MyPE;
   error("Reclaim future not implemented yet \n");
}

/* ---------------------------------------------------------------------------
**  Minim_Q()
**
**  DESCRIPTION:
**	Same as dequeue. But the smallest timestamped event is not deleted.
**	Find minimum element in the queue
**
**  DONE: MAXFLOAT here 3.4 -- HUGE_VAL
**---------------------------------------------------------------------------*/
double Minim_Q (int MyPE)
{
   int i, smallest;
   double buckettop;
   struct PEState * p_ptr;
   p_ptr = &GState[MyPE];

   buckettop = p_ptr->BucketTop;	/* copy to local var */

   if(p_ptr->Qsize == 0)
	{
     	return (MAXFLOAT);
   	}

   for(i = p_ptr->LastBucket;;) 
	{
      	if( (p_ptr->Bucket[i].Next != NULL)
	   &&
	     (p_ptr->Bucket[i].Next->Msg.RcvTS < buckettop))
		{
	 	return (p_ptr->Bucket[i].Next->Msg.RcvTS);
      		} 
	else 
		{
	 	++i;
	 	if(i == p_ptr->Nbuckets)
	    		i = 0;
	 	buckettop = buckettop + p_ptr->Width;
	 	if(i == p_ptr->LastBucket)
	    		break;
      		}
   	}

   /* Start Direct Search */
   for(smallest = 0; p_ptr->Bucket[smallest].Next == NULL; smallest++)
      	if(smallest >= p_ptr->Nbuckets) 
		{
	 	printf("Nbuckets = %d, Qsize = %d \n",
			p_ptr->Nbuckets, p_ptr->Qsize);
	 	error("Grave inconsistency in calendar queue");
      		}
   for(i = smallest + 1; i < p_ptr->Nbuckets; i++)
	{
      	if(p_ptr->Bucket[i].Next == NULL)
	 	continue;
      	else if(p_ptr->Bucket[i].Next->Msg.RcvTS 
		< p_ptr->Bucket[smallest].Next->Msg.RcvTS)
	  	smallest = i;
   	}

   return (p_ptr->Bucket[smallest].Next->Msg.RcvTS);
}


/* ---------------------------------------------------------------------------
**  Delete_Arbitrary()
**
**  DESCRIPTION:
**	Delete one element from the queue 
**
** DONE
**---------------------------------------------------------------------------*/
void Delete_Arbitrary ( struct Event ** q, int MyPE )
{
   struct PEState * p_ptr;
   p_ptr = &GState[MyPE];

   (*q)->Prev->Next = (*q)->Next;
   if((*q)->Next != NULL)
      	(*q)->Next->Prev = (*q)->Prev;
   (p_ptr->Qsize)--;
   (*q)->InPDS = FALSE;
   (*q)->Next = (*q)->Prev = NULL;
}

/* ---------------------------------------------------------------------------
**  Dump_CalQ()
**
**  DESCRIPTION:
**
** NOTDONE
**---------------------------------------------------------------------------*/
void Dump_CalQ( int MyPE, int Q )
{
}


