/******************************************************************************
*******************************************************************************
***                                                                         ***
***                       Time Warp Version 2.5                             ***
***         Copyright 1997 by Georgia Institute of Technology               ***
***                         Atlanta, Georgia                                ***
***                                                                         ***
***                      All Rights Reserved                                ***

     Permission to use, copy and modify this software and its
     documentation  for  any research  purpose and  without fee is hereby
     granted, provided that the  above copyright  notice appears in all
     copies and  that both  that  copyright  notice  and  this permission
     notice appear in supporting  documentation.  The use or inclusion of
     this  software  or its documentation in  any  commercial  product or
     distribution of this software to any other party without specific,
     written prior permission is prohibited.

     THE SOFTWARE IS PROVIDED AS IS AND GEORGIA INSTITUTE OF TECHNOLOGY
     DISCLAIMS  ALL  WARRANTIES WITH  REGARD TO THIS SOFTWARE,  INCLUDING
     ALL IMPLIED WARRANTIES  OF  MERCHANTABILITY AND FITNESS.  IN NO EVENT
     SHALL GEORGIA INSTITUTE OF TECHNOLOGY BE LIABLE FOR ANY SPECIAL,
     INDIRECT  OR CONSEQUENTIAL DAMAGES  OR ANY DAMAGES WHATSOEVER
     RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
     CONTRACT, NEGLIGENCE OR OTHER TORTIOUS  ACTION, ARISING OUT OF  OR
     IN  CONNECTION  WITH  THE  USE  OR PERFORMANCE OF THIS SOFTWARE.

***.........................................................................***
*******************************************************************************
******************************************************************************/


/*****************************************************************************
**  	TIME WARP kernel on Distributed & Shared Memory Architectures.
******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "gtw_glob.h"
#include "gtw_rand.h"
#include "gtw_utility.h"

static TW_SHARED FILE *RandFile = NULL;
TW_LOCK_TYPE RandFileLock;

/*-----------------------------------------------------------------------------
**  TWRandInit()
**
**  DESCRIPTION:  Initialize thread's private seed.  Call once before 
**	generating RNGs with successive calls to RandUnif().
**	Reads seeds from file.
**
**----------------------------------------------------------------------------*/
void TWRandInit( TWSeed *S, int skip )
{
   int i;
   long s1, s2;
   char TWSeedFile[256];
   char *gtw_root;

   if(skip < 0)
      error("TWRandInit: invalid skip parameter");

   if( (gtw_root = getenv( "GTW_ROOT" ) ) == NULL )
      {
      error("TWRandInit: GTW_ROOT environment variable not set");
      }

   sprintf( TWSeedFile, "%s/config/gtw-seeds", gtw_root );

   if( RandFile == NULL )
	{
      	RandFile = fopen(TWSeedFile, "r");
      	if(RandFile == NULL)
	    error("Couldn't open random number seed file; check installation");
   	}

   for( i = 0; i < skip; i++ )
      if( fscanf(RandFile, "%ld %ld", &s1, &s2) == EOF )
	 	{
	 	printf("EOF on GTW seed file: %s encountered \n", TWSeedFile );
         	fflush(stdout);
         	exit(1);
         	}

   if( fscanf( RandFile, "%ld %ld", &s1, &s2) == EOF )
      {
      printf("EOF on GTW seed file: %s encountered \n", TWSeedFile );
      exit(1);      
      }
   S->S1 = s1; S->S2 = s2;
}




/*-----------------------------------------------------------------------------
**  TWRandClose()
**
**  DESCRIPTION:  Initialize thread's private seed.  Call once before 
**	generating RNGs with successive calls to RandUnif().
**	Reads seeds from file.
**
**----------------------------------------------------------------------------*/
void TWRandTerminate ( void )
{

   if( RandFile == NULL ) 
	return;
   else
	{
	fclose ( RandFile );
	RandFile=NULL;
	}
}


/*-----------------------------------------------------------------------------
**  TWRandSetSeeds()
**
**  DESCRIPTION:  Initialize random number seeds to use specified value.
**
**----------------------------------------------------------------------------*/
void TWRandSetSeeds( TWSeed *S, long S1, long S2)
{
   S->S1 = S1;
   S->S2 = S2;
}


/*-----------------------------------------------------------------------------
**  TWRandGetSeeds()
**
**  DESCRIPTION:  Get random number generator seeds.
**
**----------------------------------------------------------------------------*/
void TWRandGetSeeds( TWSeed *S, long *S1, long *S2)
{
   *S1 = S->S1;
   *S2 = S->S2;
}


/*-----------------------------------------------------------------------------
**  TWRandUnif()
**
**  DESCRIPTION:  Returns a uniform random variate between 0.0 and 1.0.
**	Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
double TWRandUnif( TWSeed *S )
{
  long z, k;			/* random integer and
				   quotient respectively */

   if((S->S1 == 0) && (S->S2 == 0))
	{
        error("Random number seeds are zero:initialized?");
	}
   k = S->S1 / Q1;
   S->S1 = A1 * (S->S1 - k * Q1) - k * R1;
   if(S->S1 < 0)
      S->S1 = S->S1 + M1;
   k = S->S2 / Q2;
   S->S2 = A2 * (S->S2 - k * Q2) - k * R2;
   if(S->S2 < 0)
      S->S2 = S->S2 + M2;
   z = S->S1 - S->S2;
   if(z < 1)
      z = z + M1_MINUS_1;
   return((double) z * NORMALIZE);
}


/*-----------------------------------------------------------------------------
**  TWRandInteger()
**
**  DESCRIPTION:  Returns a uniformly distributed integer between low and
**	high inclusive.
**	Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
long TWRandInteger(TWSeed *S, long low, long high)
{
   if(high < low)
      return(0);
   else
      return(low + (long) (TWRandUnif(S) * (high + 1 - low)));

}

/*-----------------------------------------------------------------------------
**  TWRandBinomial()
**
**  DESCRIPTION:  Returns number of trials until N successes give probability P.
**      high inclusive.
**      Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
long TWRandBinomial(TWSeed *S, long N, double P)
{
   long sucesses, trials;

   sucesses = 0;

   for(trials = 0; trials < N; trials++) {
      if(TWRandUnif(S) <= P)
	 sucesses++;
   }

   return(sucesses);
}


/*-----------------------------------------------------------------------------
**  TWRandExponential ()
**
**  DESCRIPTION:  Returns a Exponential random variable with mean Lambda
**      Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
double TWRandExponential(TWSeed *S, double Lambda)
{
   return(-Lambda * log(TWRandUnif(S)));
}


/*-----------------------------------------------------------------------------
**  TWRandGamma()
**
**  DESCRIPTION:  Returns Gamma( shape, scale ).
**      high inclusive.
**      Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
double TWRandGamma(TWSeed *S, double shape, double scale)
{
   double a, b, q, phi, d;

   if(shape > 1) {
      a = 1 / sqrt(2 * shape - 1);
      b = shape - log(4);
      q = shape + 1 / a;
      phi = 4.5;
      d = 1 + log(phi);

      while(1) {
	 double U_One = TWRandUnif(S);
	 double U_Two = TWRandUnif(S);
	 double V = a * log(U_One / (1 - U_One));
	 double Y = shape * exp(V);
	 double Z = U_One * U_One * U_Two;
	 double W = b + q * V - Y;

	 double temp1 = W + d - phi * Z;
	 double temp2 = log(Z);

	 if(temp1 >= 0 || W >= temp2)
	    return(scale * Y);

      }
   } else if(shape == 1) {
      return(TWRandExponential(S, scale));
   } else {			/* shape < 1 */
      b = (exp(1) + shape) / exp(1);

      while(1) {
	 double U_One = TWRandUnif(S);
	 double P = b * U_One;

	 if(P <= 1) {
	    double Y = pow(P, (1 / shape));
	    double U_Two = TWRandUnif(S);

	    if(U_Two <= exp(-Y))
	       return(scale * Y);
	 } else {
	    double Y = -log((b - P) / shape);
	    double U_Two = TWRandUnif(S);

	    if(U_Two <= pow(Y, (shape - 1)))
	       return (scale * Y);
	 }
      }
   }
}

/*-----------------------------------------------------------------------------
**  TWRandGeometric()
**
**  DESCRIPTION:  Returns the count until the first sucess with probability P.
**      high inclusive.
**      Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
long TWRandGeometric(TWSeed *S, double P)
{
   int count = 1;

   while(TWRandUnif(S) > P)
      count++;

   return (count);
}

/*-----------------------------------------------------------------------------
**  TWRandNormal01()
**
**  DESCRIPTION:  Returns a Normal(0,1) random variable using Box-Muller Method
**      Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
double TWRandNormal01(TWSeed *S)
{
   static int FlipFlop = 0;
   static double u1, u2;

   FlipFlop = !FlipFlop;
   if((FlipFlop) || (u1 < 0.0) || (u1 >= 1.0) || (u2 < 0.0) || (u2 > 1.0)) 
	{
      	u1 = TWRandUnif(S);
      	u2 = TWRandUnif(S);
      	return (sqrt(-2.0 * log(u1)) * sin(TWOPI * u2));
   	} 
   else 
	{
      	return (sqrt(-2.0 * log(u1)) * cos(TWOPI * u2));
   	}
}

/*-----------------------------------------------------------------------------
**  TWRandNormalSD()
**
**  DESCRIPTION:   Returns a Normal(Mu, Sd) random variable with mean Mu
**	and standard deviation Sd using Box-Muller Method.
**      Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
double TWRandNormalSD(TWSeed *S, double Mu, double Sd)
{
   return (TWRandNormal01(S) * Sd + Mu);
}

/*-----------------------------------------------------------------------------
**  TWRandPoisson()
**
**  DESCRIPTION:  Returns a Poisson random variable with mean Lambda.
**      Make sure you TWRandInit() to set the seed.
**
**----------------------------------------------------------------------------*/
long TWRandPoisson(TWSeed *S, double Lambda)
{
   double a, b;
   long count;

   a = exp(-Lambda);
   b = 1;
   count = 0;

   b = b * TWRandUnif(S);

   while(b >= a) {
      count++;
      b = b * TWRandUnif(S);
   }

   return(count);
}
